#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/stat.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spinlock.h>
#include <net/protocol.h>
#include <net/pkt_sched.h>
#include <net/ip.h>

MODULE_DESCRIPTION("My kernel module");
MODULE_AUTHOR("Saurabh Jain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

#define MAX_DEVICES 8

long double MAX(long double a1,long double a2)
{
	if(a1 > a2) return a1;
	else return a2;
}

long double MIN(long double a1,long double a2)
{
	if(a1 < a2) return a1;
	else return a1;
}

long double ABS(long double a1)
{
	if(a1 < 0) return (-1)*a1;
	else return a1;
}


struct maxnethdr {
  uint8_t proto;
  float forward_price;	//For itself
  float reverse_price;	//To be feedback to the sender
} maxnethdr_t;


static unsigned long CAPACITY[] = {
				   12300,
				   12300,
				   12300,
				   12300,
				   12300,
				   12300,
				   12300,
				   12300
				  }; //B/msec
static int count = 8;
static int PROTOCOL = 203;
static int MTU =1514;
static int INTERVAL = 10; //msec
static int EFFICIENCY = 96; 

module_param_array(CAPACITY,ulong,&count,S_IRUGO|S_IWUSR);
module_param(PROTOCOL,int,S_IRUGO);
module_param(MTU,int,S_IRUGO);
module_param(INTERVAL,int,S_IRUGO);
module_param(EFFICIENCY,int,S_IRUGO);

struct maxnet_router_data {
	long double input_traffic_;
	float price_;
	unsigned long capacity_;
	uint32_t min_queue_size_;
	uint32_t max_queue_size_;
	uint8_t is_control_timer_up_;
	uint32_t control_timeout_;
	struct timer_list control_timer_;
	rwlock_t lk;
} maxnet_router_data_t;

struct nf_hook_ops netfilter_router_out;
struct maxnet_router_data maxnet_router_priv[MAX_DEVICES];

unsigned int hash_name(char * name) {
	if(name[0]=='l') return 0;
	if(name[0]=='e' && name[3]=='0') return 1;
	if(name[0]=='e' && name[3]=='1') return 2;
	if(name[0]=='e' && name[3]=='2') return 3;
	if(name[0]=='e' && name[3]=='3') return 4;
	if(name[0]=='e' && name[3]=='4') return 5;
	if(name[0]=='e' && name[3]=='5') return 6;
	if(name[0]=='s' && name[3]=='0') return 7;
	return 0;
}


void control_timeout_handler(unsigned long data)
{
	int cpu;
	long double input_bw = 0;
	struct maxnet_router_data *str = (struct maxnet_router_data*) data;
	if(str==NULL) {
		printk(KERN_CRIT "#timeout handler: str cannot be null.\n");
		return;
	}
	cpu = get_cpu();
	write_lock(&(str->lk));
	del_timer_sync(&(str->control_timer_));
	str->is_control_timer_up_ = 0;
	input_bw = (str->input_traffic_)/(str->control_timeout_);
	str->price_ = str->price_ + ((float)(input_bw - EFFICIENCY*str->capacity_))/((float)(str->capacity_));
			
	if(printk_ratelimit())
	{
		printk(KERN_INFO "Router-ctimeout:");
		printk(KERN_INFO "%u:",(unsigned int)(str->capacity_));
		printk(KERN_INFO "%ld:",(long int)(str->input_traffic_));
		printk(KERN_INFO "%ld:",(long int)(str->control_timeout_));
		printk(KERN_INFO "%ld:",(long int)(str->price_));
		printk(KERN_INFO "%ld:",(long int)(str->min_queue_size_));
		printk(KERN_INFO "%ld:",(long int)(str->max_queue_size_));
	}

	str->input_traffic_ = 0;
	str->min_queue_size_ = 0;
	str->max_queue_size_ = 0;
	init_timer(&(str->control_timer_));
	str->control_timer_.expires = (jiffies + (unsigned long)(((str->control_timeout_)*HZ)/1000.0));
	str->control_timer_.data = (unsigned long)str;
	str->control_timer_.function = control_timeout_handler;
	add_timer(&(str->control_timer_));
	str->is_control_timer_up_ = 1;
	write_unlock(&(str->lk));
	put_cpu();
}

unsigned int pkt_handler(unsigned int hooknum, struct sk_buff **skb,
			const struct net_device *in, const struct net_device *out,
			int (*okfn)(struct sk_buff*))
{
	unsigned int i=hash_name((char*)out->name);
	uint32_t packetsize = 0;
	uint32_t queuelen = 0; 
	uint32_t ip_header_len;
	int cpu;
	struct maxnet_router_data *str;
	struct sk_buff *buff;
	struct iphdr *ihdr;
	struct maxnethdr *xhdr;
	if(i < 0 || i > MAX_DEVICES) printk(KERN_CRIT "#pkt handler: invalid hash calculated: %d\n",i); 

	str = &maxnet_router_priv[i];
	buff = *skb;
	ihdr = (struct iphdr*)buff->nh.iph;
	ip_header_len = (ihdr->ihl) * 4;
	if(ihdr->protocol==PROTOCOL) {
		xhdr = (struct maxnethdr*)((buff->data) + ip_header_len);
		packetsize = (buff->len);
		write_lock(&(str->lk));
		cpu = get_cpu();
		str->input_traffic_ +=packetsize;
		if(str->price_ > xhdr->forward_price) {
			xhdr->forward_price = str->price_;
		}
		put_cpu();

		if((out != 0) && (out->qdisc != 0)) {
			queuelen = out->qdisc->q.qlen;
			if(queuelen < str->min_queue_size_) str->min_queue_size_ = queuelen;
			str->max_queue_size_ = MAX(str->max_queue_size_,queuelen);
		}
		write_unlock(&(str->lk));
	}
	return NF_ACCEPT;
}


static int maxnet_r_init_module(void)
{
	int i = 0;
	netfilter_router_out.hooknum  = NF_IP_POST_ROUTING;
	netfilter_router_out.priority = NF_IP_PRI_LAST;
	nf_register_hook(&netfilter_router_out);

	for (i = 0; i < MAX_DEVICES; i++) 
	{
		rwlock_init(&(maxnet_router_priv[i].lk));
		maxnet_router_priv[i].input_traffic_ = 0;
		maxnet_router_priv[i].control_timeout_ = INTERVAL;
		maxnet_router_priv[i].min_queue_size_ = 0;	
		maxnet_router_priv[i].max_queue_size_ = 0;	
		init_timer(&(maxnet_router_priv[i].control_timer_));
		maxnet_router_priv[i].control_timer_.expires = (jiffies + (INTERVAL * HZ)/1000);
		maxnet_router_priv[i].control_timer_.data = (unsigned long)&maxnet_router_priv[i];
		maxnet_router_priv[i].control_timer_.function = control_timeout_handler;
		add_timer(&(maxnet_router_priv[i].control_timer_));
		maxnet_router_priv[i].is_control_timer_up_ = 1;
		maxnet_router_priv[i].capacity_ = CAPACITY[i];
	}
	printk( KERN_ALERT "Module MaxNet_r init\n" );
	return 0;
}

static void maxnet_r_exit_module(void)
{
	int i = 0;
	nf_unregister_hook(&netfilter_router_out);
	for (i = 0; i < MAX_DEVICES; i++) {
		if (maxnet_router_priv[i].is_control_timer_up_==1) {
			maxnet_router_priv[i].is_control_timer_up_ = 0;
			del_timer_sync(&(maxnet_router_priv[i].control_timer_));
		}
	}
	printk( KERN_ALERT "Module MaxNet_r exit\n" );
}

module_init(maxnet_r_init_module);
module_exit(maxnet_r_exit_module);
