#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/tcp.h>

MODULE_DESCRIPTION("RCP Host Module");
MODULE_AUTHOR("Saurabh Jain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

#define MAX(x,y) (x > y ? x : y)
#define MIN(x,y) (x < y ? x : y)
#define ABS(x) (x > 0 ? x : (-1)*(x))

struct rcphdr {
	float fair_rate_;
	float fair_rate_feedback_;
	float rtt_;
	float fw_intnum_;
	float bk_intnum_;
};

struct perflow {
	int hash;
	struct tcp_sock *to;
	unsigned long bw_data;
	unsigned long bw_last_update;
	int zero_data_count;
	struct timer_list bw_timer;
	int is_bw_timer_up;
	rwlock_t lk;
};

static int PROTOCOL = 204;
static struct net_protocol rcpproto;
static struct perflow flows[1024];
int (*curr_queue_xmit) (struct sk_buff* skb, int ipfragok);
int netheader_len = 0;
static int BW_INTERVAL = 100;
static int IS_RX = 0;
static int KILL_TIMER_CNT = 10;
void bw_timer_handler(unsigned long);

module_param(PROTOCOL,int,S_IRUGO);
module_param(BW_INTERVAL,int,S_IRUGO);
module_param(IS_RX,int,S_IRUGO);
module_param(KILL_TIMER_CNT,int,S_IRUGO);

extern struct net_protocol tcp_protocol;
extern struct tcp_func ipv4_specific;
//extern void tcp_rcp_output(unsigned long);

struct perflow* get_flow(uint32_t daddr, uint32_t saddr, uint16_t dport, uint16_t sport) 
{
  unsigned char c = 0;
  unsigned int h = (saddr ^ sport) ^ (daddr ^ dport);
  h ^= h >> 16;
  h ^= h >> 8;
  c = (unsigned char)h;
  flows[c].hash = c;
  return &(flows[c]);
}

//IP->RCP->TCP
//Possible Cases: 1) Receiver: receives a Data packet 
//                2) Sender: receives an Ack packet
//                3) Sender: receives a Data-Ack packet
int rcp_pkt_handler(struct sk_buff* skb) 
{
	struct perflow* pd = 0;
	struct rcphdr* hdr = (struct rcphdr*)skb->data;
	float fair_rate;
	float fair_rate_feedback;
	float fw_intnum_ = hdr->fw_intnum_;
	float bk_intnum_ = hdr->bk_intnum_;
	fair_rate = hdr->fair_rate_;
	fair_rate_feedback = hdr->fair_rate_feedback_;
	

	//Remove header
	skb_pull(skb,sizeof(struct rcphdr));
	skb->h.th = (struct tcphdr*)skb->data;
	
	pd = get_flow(skb->nh.iph->daddr,skb->nh.iph->saddr,
		ntohs((short)(skb->h.th->dest)),
		ntohs((short)(skb->h.th->source)));

	if(pd == NULL) {
		printk(KERN_CRIT "#rcp_pkt_handler: Null value of pd obtained.\n");
		return 0;
	}

	if(pd->to != NULL) {
		pd->to->rcp.netheader_len = netheader_len;
		pd->to->rcp.fair_rate_feedback = fair_rate;
		pd->to->rcp.rate_ = (int)fair_rate_feedback;
		pd->to->rcp.fw_intnum_ = bk_intnum_;
		pd->to->rcp.bk_intnum_ = fw_intnum_;
		pd->to->rcp.bytes_per_send_ = (int)fair_rate_feedback;

		//del_timer_sync(&(pd->to->rcp.send_timer_));
		//if(pd->to->rcp.active_ == 1) {
		//	init_timer(&(pd->to->rcp.send_timer_));
		//	pd->to->rcp.send_timer_.expires = (jiffies+1);
		//	pd->to->rcp.send_timer_.data = (unsigned long)(pd->to);
		//	pd->to->rcp.send_timer_.function = tcp_rcp_output;
		//	add_timer(&(pd->to->rcp.send_timer_));
		//	pd->to->rcp.istimer_up_ = 1;
		//}

		//printk(KERN_INFO "%d\n",(int)pd->to->rcp.rate_);
	}
	return tcp_protocol.handler(skb);
}

//TCP->RCP->IP
//Possible cases: 1) Sender: Sending a data packet 
//		  2) Receiver: Sending an Ack 
//		  3) Receiver: Sending a Data-Ack packet
int rcp_to_ip_pkt_dispatcher(struct sk_buff* skb, int ipfragok) 
{
	int len = 0;
	struct rcphdr* hdr = 0;
	struct perflow* pd = 0;
	struct inet_sock* io = inet_sk(skb->sk);
	struct tcphdr* thdr = (struct tcphdr*)skb->h.th;
	struct tcp_skb_cb *tcb = TCP_SKB_CB(skb);
	len = skb->len;

	pd = get_flow(io->daddr,io->saddr,ntohs((short)(thdr->dest)),
		ntohs((short)(thdr->source)));

	if(pd==NULL) {
		printk(KERN_CRIT "#Host-dispatcher: Null value of pd obtained.\n");
		return 0;
	}

	if((tcb->flags & TCPCB_FLAG_SYN))
	{
		pd->to = NULL;
	}

	pd->to = tcp_sk(skb->sk);
	pd->to->rcv_wnd = 107374182;
	pd->to->rcv_ssthresh = 107374182;
	hdr = (struct rcphdr*)skb_push(skb,sizeof(struct rcphdr));
	hdr->fair_rate_ = -1;
	hdr->fair_rate_feedback_ = 0;
	hdr->rtt_ = 1;
	hdr->fw_intnum_ = 0;
	hdr->bk_intnum_ = 0;
	skb->sk->sk_protocol = PROTOCOL;

	if(pd != NULL && pd->to != NULL) {
		hdr->fair_rate_feedback_ = pd->to->rcp.fair_rate_feedback;
		hdr->rtt_ = (pd->to->srtt)/8;
		if(hdr->rtt_ == 0) hdr->rtt_ = 1;
		hdr->fw_intnum_ = pd->to->rcp.fw_intnum_;
		hdr->bk_intnum_ = pd->to->rcp.bk_intnum_;
		if(pd->to->rcp.active_ == 1 && IS_RX == 0) {
			if(pd->is_bw_timer_up == 0 && pd->zero_data_count <= KILL_TIMER_CNT)  {
				pd->bw_last_update = jiffies;
				pd->bw_data = (len + netheader_len);
				init_timer(&(pd->bw_timer));
				pd->bw_timer.data = (unsigned long)pd;
				pd->bw_timer.function = bw_timer_handler;
				pd->bw_timer.expires = jiffies + (BW_INTERVAL*HZ)/1000;
				add_timer(&(pd->bw_timer));
				pd->is_bw_timer_up = 1;
			} else {
				write_lock(&(pd->lk));
				pd->bw_data += (len + netheader_len);
				write_unlock(&(pd->lk));
			}
		}
	}
	return curr_queue_xmit(skb,ipfragok);
}

void bw_timer_handler(unsigned long pd) {
	int interval = 0;
	unsigned long traffic = 0;
	struct perflow *str = (struct perflow*)pd;
	del_timer_sync(&(str->bw_timer));
	write_lock(&(str->lk));
	traffic = str->bw_data;

	if(traffic == 0)
		str->zero_data_count++;
	else
		str->zero_data_count = 0;

	str->bw_data = 0;
	interval = jiffies - str->bw_last_update;
	str->bw_last_update = jiffies;
	write_unlock(&(str->lk));

	if(interval > 0) {
		/*
		printk(KERN_CRIT "%d:%ld:%ld:%d:%d:%d:%d\n",str->hash,str->bw_last_update,
					(unsigned long)traffic/interval,(int)str->to->rcp.rate_,
					interval,str->to->snd_cwnd,str->to->srtt/8);
		*/
	} else {
		printk(KERN_CRIT "#bw_timer_handler: interval is <= 0");
	}

 	str->bw_data = 0;

	if(str->to->rcp.active_ == 1 && IS_RX == 0 && 
				str->zero_data_count <= KILL_TIMER_CNT) {

		init_timer(&(str->bw_timer));
		str->bw_timer.data = (unsigned long)pd;
		str->bw_timer.function = bw_timer_handler;
		str->bw_timer.expires = str->bw_last_update + (BW_INTERVAL*HZ)/1000;
		add_timer(&(str->bw_timer));
	}
}

void rcp_err (struct sk_buff* skb, u32 info)
{
	//printk("#RCP error handler called : %u\n",info);
	tcp_protocol.err_handler(skb,info);
}

static int rcp_host_init_module(void)
{
	int i = 0;
	struct tcp_func* tf = &ipv4_specific;
	rcpproto.handler = rcp_pkt_handler;
	rcpproto.err_handler = rcp_err;
	rcpproto.no_policy = 1;
	inet_add_protocol(&rcpproto,PROTOCOL);
	curr_queue_xmit = tf->queue_xmit;
	tf->queue_xmit = rcp_to_ip_pkt_dispatcher;
	tf->net_header_len += sizeof(struct rcphdr);
	netheader_len = tf->net_header_len;
	
	for (i = 0; i < 1024; i++)
	{
		struct perflow* a = &(flows[i]);
		a->hash = i;
		a->to = NULL;
	}
	printk( KERN_INFO "Module rcp_host init\n" );
	return 0;
}

static void rcp_host_exit_module(void)
{
	struct tcp_func* tf = &ipv4_specific;	
	inet_del_protocol(&rcpproto,PROTOCOL);	
	tf->queue_xmit = curr_queue_xmit;
	tf->net_header_len -= sizeof(struct rcphdr); 
	printk( KERN_INFO "Module rcp_host exit\n" );
}

module_init(rcp_host_init_module);
module_exit(rcp_host_exit_module);
