#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/jiffies.h>


MODULE_DESCRIPTION("XCP Host");
MODULE_AUTHOR("saujain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

#define MAX(x,y) (x > y ? x : y)
#define MIN(x,y) (x < y ? x : y)
#define ABS(x) (x > 0 ? x : (-1)*(x))

struct xcphdr {
  uint8_t proto;
  uint8_t len;
  uint8_t verform;
  uint8_t unused;
  float rtt;  
  float x;
  float delta;
  float reverse;
};

struct perflow {
  int hash;
  struct tcp_sock *to;
  float increment; //for you (B/msec)
  float reverse; //for remote user (B/msec)
};

static unsigned long CAPACITY = 125000; //B/msec
static int PROTOCOL = 201;
static int MTU = 1514;
static unsigned long DESIRED_RATE=125000; //B/msec

static struct net_protocol xcpproto;
static struct perflow flows[1024];
int (*curr_queue_xmit) (struct sk_buff* skb, int ipfragok);

module_param(CAPACITY,ulong,S_IRUGO);
module_param(DESIRED_RATE,ulong,S_IRUGO);
module_param(PROTOCOL,int,S_IRUGO);
module_param(MTU,int,S_IRUGO);

extern struct net_protocol tcp_protocol;
extern struct tcp_func ipv4_specific;

struct perflow* get_flow(uint32_t daddr, uint32_t saddr, uint16_t dport, uint16_t sport) 
{
  unsigned char c = 0;
  unsigned int h = (saddr ^ sport) ^ (daddr ^ dport);
  h ^= h >> 16;
  h ^= h >> 8;
  c = (unsigned char)h;
  flows[c].hash = c;
  return &(flows[c]);
}

//IP->XCP->TCP
//Possible Cases: 1) Receiver: receives a Data packet 
//                2) Sender: receives an Ack packet
//                3) Sender: receives a Data-Ack packet
int xcp_pkt_handler(struct sk_buff* skb) 
{
  float delta;
  float reverse;
  float rtt = 0;
  struct perflow* pd = 0;
  struct xcphdr* hdr = (struct xcphdr*)skb->data;
  int cpu;
  long int increment = 0, max=0;
  delta = hdr->delta;
  reverse = hdr->reverse;
  
  if ((((hdr->proto) & (MAX_INET_PROTOS -1))) != 6) 
  {
    printk(KERN_CRIT "#xcp_pkt_handler: Unknown protocol to call next.\n");
    return 0;
  } 

  skb_pull(skb,hdr->len);
  skb->h.th = (struct tcphdr*)skb->data;
    
  pd = get_flow(skb->nh.iph->daddr,skb->nh.iph->saddr,
		ntohs((short)(skb->h.th->dest)),
		ntohs((short)(skb->h.th->source)));
		

  if(pd==NULL) {
	printk(KERN_CRIT "#xcp_pkt_handler: Null value of pd obtained.\n");
	return 0;
  }
  
  cpu = get_cpu();
  pd->reverse += delta;
  pd->increment += reverse;

  if(pd->to!=NULL) {
	  pd->to->snd_cwnd_cnt = 0;
  }

  put_cpu();

  
  if(pd->to!=NULL && (pd->to->srtt) !=0 && pd->increment!=0) {
  	cpu = get_cpu();
  	rtt = ((float)(pd->to->srtt))/8.0;
	if(pd->to->ca_state == TCP_CA_Open) 
	{ 
  	increment = (long int)(((pd->increment)*rtt)/(MTU));

	if(increment >=0) {
		max = (pd->to->snd_cwnd + increment);
	  if(max > (long int)((pd->to->snd_cwnd_clamp))) {
        increment = (pd->to->snd_cwnd_clamp - pd->to->snd_cwnd);
    }
    pd->to->snd_cwnd = (pd->to->snd_cwnd + increment);
		
		if(increment != 0) {
			printk(KERN_INFO "%d:%ld:%d:%ld:%ld:%d\n",pd->hash,
      	                                jiffies,(int)rtt,
                                      (long)(pd->to->snd_cwnd),
										      (long)(tcp_packets_in_flight(pd->to)),	
                                      (int)(pd->to->ca_state));
		}

	} else {
		if(ABS(increment) >= (long int)(pd->to->snd_cwnd)) {
                        pd->to->snd_cwnd = 1;
			increment = 0;
			pd->increment = 0;
                } else {
                        pd->to->snd_cwnd = (pd->to->snd_cwnd + increment);
                }
		printk(KERN_INFO "%d:%ld:%d:%ld:%ld:%d\n",pd->hash,
                                      jiffies,(int)rtt,
				      (long)(pd->to->snd_cwnd),
				      (long)(tcp_packets_in_flight(pd->to)),
				      (int)(pd->to->ca_state));
	}
	pd->increment = (pd->increment - (increment*MTU)/rtt);
	pd->to->snd_ssthresh = (pd->to->snd_cwnd)/2;
  pd->to->snd_cwnd_stamp = tcp_time_stamp;
	} 
	else {
		pd->increment = 0;
		printk(KERN_INFO "%d:%ld:%d:%ld:%ld:%d\n",pd->hash,
                                      jiffies,(int)rtt,
                                      (long)(pd->to->snd_cwnd),
				      (long)(tcp_packets_in_flight(pd->to)),
                                      (int)(pd->to->ca_state));
	}
  put_cpu();
  }
  

  return tcp_protocol.handler(skb);
}

//TCP->XCP->IP
//Possible cases: 1) Sender: Sending a data packet 
//		  2) Receiver: Sending an Ack 
//		  3) Receiver: Sending a Data-Ack packet
int xcp_to_ip_pkt_dispatcher(struct sk_buff* skb, int ipfragok) 
{
  int cpu = 0;
	int len = 0;
  float throughput = 0;
  struct xcphdr* hdr = 0;
  struct perflow* pd = 0;
  struct inet_sock* io = inet_sk(skb->sk);
  struct tcphdr* thdr = (struct tcphdr*)skb->h.th;
	struct tcp_skb_cb *tcb = TCP_SKB_CB(skb);
  len = skb->len;
  
  pd = get_flow(io->daddr,io->saddr,ntohs((short)(thdr->dest)),
		ntohs((short)(thdr->source)));

  if(pd==NULL) {
      printk(KERN_CRIT "#Host-dispatcher: Null value of pd obtained.\n");
      return 0;
  }

	if((tcb->flags & TCPCB_FLAG_SYN))
	{
		pd->reverse = 0;
		pd->to = NULL;
	}

  pd->to = tcp_sk(skb->sk);
	pd->to->rcv_wnd = 107374182;
	pd->to->rcv_ssthresh = 107374182;
  cpu = get_cpu();
  hdr = (struct xcphdr*) skb_push(skb,sizeof(struct xcphdr));
  hdr->proto = 6;
  hdr->len = sizeof(struct xcphdr);
  hdr->unused = 0;
  hdr->verform = 0x11;
  hdr->rtt = ((float)(pd->to->srtt))/8.0;
  hdr->x = 0;
  hdr->delta = 0;
  hdr->reverse = pd->reverse;
  pd->reverse = 0;

  skb->sk->sk_protocol = PROTOCOL;
    
  if (skb->len == ((4*thdr->doff) + sizeof(struct xcphdr))) {
 		hdr->rtt	 = 0;	
  }
 
	else if(hdr->rtt !=0 && pd->to->snd_cwnd !=0) {
	
		hdr->x = (hdr->rtt)/(pd->to->snd_cwnd);
		throughput = (pd->to->snd_cwnd*MTU)/(hdr->rtt);
	
		if(throughput >= DESIRED_RATE) 
		{
			hdr->delta = 0;
		} else {
			hdr->delta = ((DESIRED_RATE - throughput)*MTU)/(throughput*hdr->rtt);
		}

  }
	put_cpu();
  return curr_queue_xmit(skb,ipfragok);
}    

void xcp_err (struct sk_buff* skb, u32 info) 
{
  //printk("#XPC error handler called : %u\n",info);
  tcp_protocol.err_handler(skb,info);
}

static int xcp_host_init_module(void)
{
	int i =0;
	struct tcp_func* tf = &ipv4_specific;
	xcpproto.handler = xcp_pkt_handler;
	xcpproto.err_handler = xcp_err;
	xcpproto.no_policy = 1;
	inet_add_protocol(&xcpproto,PROTOCOL);
	curr_queue_xmit = tf->queue_xmit;
	tf->queue_xmit = xcp_to_ip_pkt_dispatcher;
	tf->net_header_len += sizeof(struct xcphdr);	
	
	for (i = 0; i < 1024; i++) 
	{
		struct perflow* a = &(flows[i]);
		a->hash = i;
		a->reverse = 0;
		a->to = NULL;
	}
	printk(KERN_ALERT "#Module xcp_host init\n");
	return 0;
}

static void xcp_host_exit_module(void)
{
	struct tcp_func* tf = &ipv4_specific;	
  	inet_del_protocol(&xcpproto,PROTOCOL);	
  	tf->queue_xmit = curr_queue_xmit;
  	tf->net_header_len -= sizeof(struct xcphdr); 
  	printk(KERN_ALERT "#Module xcp_host exit done! \n");
}

module_init(xcp_host_init_module);
module_exit(xcp_host_exit_module);
